import os
import pandas as pd
import geopandas as gpd
import numpy as np
import xarray as xr
from rasterstats import zonal_stats
from shapely import wkt
import copy
import requests
import tempfile
from io import BytesIO
from zipfile import ZipFile
from urllib.request import urlopen
from shapely.ops import transform

from . import parameters as pr
from .utilib import download_re_ninja
from .demand import Power, Heat
from .renewables import Heat_Pumps, VRE_Capacity_Factor, Hydro, Area
from .resources import Power_Plants
from .classification import disaggregation, aggregation, round_coord, max_p_regions
from .model import create_location_yaml, create_timeseries_csv, create_model_yaml

class EUSES():

    def __init__(self,countries,year,import_ds=False,temperature_weighting='population'):
        '''
        countries : List of countries by name
        year : reference year
        '''

        self.ds = xr.Dataset()
        self.countries = countries
        self.year = year

        if import_ds == False:
            print('Building Areas dataset')

            nuts_geom_eu = gpd.read_file('https://gisco-services.ec.europa.eu/distribution/v2/nuts/geojson/NUTS_RG_10M_2013_3035_LEVL_2.geojson')

            time = pd.date_range(str(year), str(year+1), freq='1H')[:-1]
            self.ds.coords['time'] = time
            self.ds.coords['nuts_0'] = [pr.get_metadata(c,'nuts_id') for c in self.countries]

            nuts_2 = gpd.GeoDataFrame()
            for c in self.countries:
                nuts_id = pr.get_metadata(c,'nuts_id')
                if nuts_id == 'GR':
                    nuts_geom = nuts_geom_eu.loc[nuts_geom_eu['id'].str.contains('EL')]
                elif nuts_id == 'GB':
                    nuts_geom = nuts_geom_eu.loc[nuts_geom_eu['id'].str.contains('UK')]
                else:
                    nuts_geom = nuts_geom_eu.loc[nuts_geom_eu['id'].str.contains(nuts_id)]

                nuts_2 = gpd.GeoDataFrame(pd.concat([nuts_geom, nuts_2]))

            nuts_2.crs = {'init': 'epsg:3035'}
            # remove islands from nuts zones
            islands = ['FRA','FR83','PT20','PT30','ES53','ES63','ES64','ES70',
                                'EL30','EL41','EL65','EL43','EL62','EL42']
            for id in islands:
                nuts_2 = nuts_2.loc[nuts_2['id'].str.contains(id)==False]

            self.ds.coords['nuts_2'] = (('nuts_2'), nuts_2['NUTS_ID'])
            self.ds['country_code'] = ('nuts_2'),(nuts_2['CNTR_CODE'].to_xarray().astype('str'))
            self.ds['geometry'] = (('nuts_2'),(nuts_2['geometry'].to_xarray()))
            self.ds['geometry_54009'] = (('nuts_2'),(nuts_2['geometry'].to_crs({'proj': 'moll'}).to_xarray()))
            self.ds['geometry'].attrs['crs'] = 'epsg:3035'

            print('  - {} added'.format('NUTS 2 ID and geometry'))
            eu_gdp = pd.read_csv('data/input_data/nuts2_gdp.csv',header=2,index_col=0)
            self.ds['gdp'] = eu_gdp.loc[self.ds.coords['nuts_2'].values,'2015']
            self.ds['gdp'] = self.ds['gdp'].fillna(0)

            # add population data
            if ['population'] not in list(self.ds.keys()):
                temp = tempfile.TemporaryDirectory()
                resp = urlopen("http://cidportal.jrc.ec.europa.eu/ftp/jrc-opendata/GHSL/GHS_POP_GPW4_GLOBE_R2015A/GHS_POP_GPW42015_GLOBE_R2015A_54009_1k/V1-0/GHS_POP_GPW42015_GLOBE_R2015A_54009_1k_v1_0.zip")
                zipfile = ZipFile(BytesIO(resp.read()))
                raster_path = temp.name+'/GHS_POP_GPW42015_GLOBE_R2015A_54009_1k_v1_0.tif'
                open(raster_path, 'wb').write(zipfile.read(zipfile.namelist()[3]))
                population_data = []
                for l in  self.ds.coords['nuts_2']:
                    geo = self.ds.sel(nuts_2=l)['geometry_54009'].values.item()
                    population = zonal_stats(geo, raster_path, stats='sum')[0].get('sum')
                    population_data.append(population)

                self.ds['population'] = (('nuts_2'),(np.array(population_data)))
                self.ds['population'].attrs['unit'] = 'People'
                self.ds = self.ds.drop('geometry_54009')
                temp.cleanup()
            print('  - {} variables added'.format('Population and gdp'))

            # add temperature data
            r = urlopen("https://zenodo.org/record/4584576/files/era-nuts-t2m-nuts2-hourly.nc")
            ds_t2m = xr.open_dataset(BytesIO(r.read())).load()
            diff = list(set(self.ds.nuts_2.values).difference(set(ds_t2m['t2m'].region.values)))

            nuts2_13_16 = pd.read_excel("https://ec.europa.eu/eurostat/documents/345175/629341/NUTS2013-NUTS2016.xlsx",engine='openpyxl',sheet_name='Correspondence NUTS-2',index_col=0,header=0)
            corr = nuts2_13_16.loc[list(set(diff).intersection(set(nuts2_13_16.index))),'Code 2016'].dropna()

            self.ds['temperature'] = (('nuts_2','time'),(np.empty((len(self.ds.nuts_2),len(self.ds.time)))))
            common_nuts2 = list(set(self.ds.nuts_2.values).intersection(set(ds_t2m['t2m'].region.values)))
            self.ds['temperature'].loc[common_nuts2,:] = ds_t2m['t2m'].loc[self.ds.time,common_nuts2].transpose().values
            self.ds['temperature'].loc[corr.index,:] = ds_t2m['t2m'].loc[self.ds.time,corr.values].transpose().values

            realocated = {'HU10':['HU11','HU12'],'LT00':['LT01','LT02'],'PL12':['PL91','PL91'],'UKM3':['UKM9','UKM8'],
            'IE02':['IE05','IE06'],'IE01':['IE04']}
            for code_2013, code_2016 in realocated.items():
                if code_2013 in self.ds.nuts_2.values:
                    self.ds['temperature'].loc[code_2013,:] = ds_t2m['t2m'].loc[self.ds.time,code_2016].mean(axis=1).transpose().values
            print('  - {} variables added'.format('Temperature'))

    def add(self, component,  **kwargs):
        comp_class = eval(component)
        comp_class(self, **kwargs)

    def save_dataset(self, dir_name):

        if os.path.exists("data/saved_dataset/" + dir_name):
            os.remove("data/saved_dataset/" +  dir_name)
        geo_list = [str(geo) for geo in self.ds['geometry'].values]
        self.ds['geometry'] = (('nuts_2'),(geo_list))
        encoding = {k: {'zlib': True, 'shuffle': True} for k in self.ds.variables}
        self.ds.to_netcdf("data/saved_dataset/" + dir_name, encoding=encoding)
        self.ds['geometry'] = (('nuts_2'),pd.Series(self.ds['geometry']).apply(wkt.loads))

    def create_regions(self, method, area_factor=None, initial_val=1, initial_seed=1,disaggr_var='preset', aggr_var ='preset'):

        ds = self.ds.copy(deep=True)
        disaggregation(ds,disaggr_var) # add wind-offshore capacity factor to nuts_2 areas

        island_groups = [['DK01','DK02','DK03'],['FI20','FI1B'],['ITG2','ITG1','ITF6'],['UKM3','UKN0'], ]

        ds = aggregation(ds, island_groups,aggr_var)

        if method not in ['nuts0','nuts1','nuts2']:
            if 'BE34' and 'LU00' in ds.coords['nuts_2'].values:
                ds = aggregation(ds, [['BE34','LU00']],aggr_var)

        zones = gpd.GeoDataFrame(geometry=ds['geometry'].values)
        zones['id'] = ds.coords['nuts_2'].values

        zones['pv_pot_e'] = (ds['rooftop_pv']+ds['utility_pv']).values * ds['pv_cf'].values.sum(axis=1)
        zones['wind_on_pot_e'] = ds['onshore_wind'].values * ds['wind_cf'].values.sum(axis=1)
        zones['wind_off_pot_e'] = ds['offshore_wind'].values * ds['wind_offshore_cf'].values.sum(axis=1)

        zones['flh_max']=zones.loc[zones.pv_pot_e<zones.wind_on_pot_e].wind_on_pot_e
        zones.update(pd.Series(zones.loc[zones.pv_pot_e>zones.wind_on_pot_e].pv_pot_e, name='flh_max'))
        zones.update(pd.Series(zones.loc[zones.wind_off_pot_e>zones.flh_max].wind_off_pot_e, name='flh_max'))
        zones['flh_max'] = zones['flh_max'].fillna(0)

        zones['storage'] = ds['hydro_storage'].sum(axis=1)
        zones['population'] = ds['population']

        zones['minimum_threshold'] = zones.geometry.area

        method_dir = {'rdm_regions':['rdm_values'],'max_p_regions':['population','flh_max','storage'],
                        'nuts0':2,'nuts1':3,'nuts2':4}

        if method in ['rdm_regions','max_p_regions']:
            np.random.seed(initial_seed)
            zones['rdm_values'] = np.random.rand(len(zones.index))
            feature = method_dir.get(method)
            class_regions_int = max_p_regions(zones,feature,area_factor, initial_val)
            if class_regions_int == False:
                zones['nuts'] = zones['id'].str[:2]
                nuts_array = zones['nuts'].unique()
                class_regions_int = [np.where(zones['nuts']==x)[0].tolist() for x in nuts_array]

        if method in ['nuts0','nuts1','nuts2']:
            zones['nuts'] = zones['id'].str[:method_dir.get(method)]
            nuts_array = zones['nuts'].unique()
            class_regions_int = [np.where(zones['nuts']==x)[0].tolist() for x in nuts_array]


        class_regions = [zones.loc[c].id.to_list() for c in class_regions_int]

        ds = aggregation(ds,class_regions,aggr_var)
        ds.coords['regions'] = ('nuts_2', ds.coords['nuts_2'].values)
        ds = ds.swap_dims({'nuts_2': 'regions'})
        ds = ds.drop('nuts_2')
        if type(ds['country_code'][0].values) == np.ndarray:
            country_codes = [c.item() for c in ds['country_code']]
            ds['country_code'].values = country_codes

        self.ds_regions = ds

    def create_calliope_model(self, op_mode='plan',sectors = ['power','heat'],co2_cap_factor=None, national=False):
        '''
        op_mode: either 'plan' or 'operate'
        '''
        ds_regions = self.ds_regions

        regions_geo = gpd.GeoDataFrame(columns=['geometry'], geometry=ds_regions['geometry'].values)
        regions_geo['id'] = ['region_'+str(e) for e,id in enumerate(ds_regions['geometry'].coords['regions'].values)]
        regions_geo['nuts_2s'] = ds_regions['geometry'].coords['regions'].values

        if national == True:
            regions_geo['id'] = ds_regions['country_code'].values

        regions_geo.crs = {'init': 'epsg:3035'}

        regions_geo = regions_geo.to_crs({'init': 'epsg:4326'})

        create_timeseries_csv(regions_geo, ds_regions)
        create_location_yaml(regions_geo, ds_regions,sectors)
        create_model_yaml(self, regions_geo, sectors, op_mode, co2_cap_factor)

    def filter_countries(self, countries):
        filt_ds = copy.deepcopy(self)
        [filt_ds.countries.remove(c) for c in countries]
        nuts_0s = [pr.get_metadata(c,'nuts_id') for c in countries]
        if 'EE00' in nuts_0s:
            nuts_0s.remove('EE00')
            nuts_0s.append('EE')
        if 'GR' in nuts_0s:
            nuts_0s.remove('GR')
            nuts_0s.append('EL')
        nuts_0s_invers = [pr.get_metadata(c,'nuts_id') for c in filt_ds.countries]
        nuts_2s = self.ds.where(self.ds['country_code'].isin(nuts_0s), drop = True).coords['nuts_2'].values
        nuts_2s_invers = list(self.ds.coords['nuts_2'].values)
        [nuts_2s_invers.remove(n2) for n2 in nuts_2s]
        filt_ds.ds = filt_ds.ds.drop(nuts_0s_invers,dim='nuts_0')
        filt_ds.ds = filt_ds.ds.drop(nuts_2s_invers,dim='nuts_2')
        filt_ds.countries = countries
        if type(filt_ds.ds['country_code'][0].values) == np.ndarray:
            country_codes = [c.item() for c in filt_ds.ds['country_code']]
            filt_ds.ds['country_code'].values = country_codes

        return filt_ds

def import_dataset(dir_name):
    ds = xr.open_dataset("data/saved_dataset/" + dir_name)
    countries = [pr.get_metadata(id,'name') for id in ds.coords['nuts_0'].values]
    year = pd.to_datetime(ds.coords['time'].values)[0].year
    self = EUSES(countries,year,import_ds=True)
    self.ds = ds
    self.ds['geometry'] = (('nuts_2'),pd.Series(self.ds['geometry']).apply(wkt.loads))
    return self

def build_dataset(countries, year=2010, save=True, dir_name = 'dataset.nc'):
    # Make list of all countries considered in NUTS 2 dataset
    if countries == 'EU':
        countries_metadata = pr.countries_metadata()
        countries = [country.get('name') for country in countries_metadata]

    # Build NUTS 2 dataset in EUSES dataset for the year

    self = EUSES(countries, year)
    # Add data components
    data_components_list = {'Power_Plants':'Power plants',
                            'Area':'Onshore wind, offshore wind, and solar PV area',
                            'Hydro':'Hydropower capacity and availability',
                            'Heat_Pumps': 'Air-sourced HP capacity factor',
                            'VRE_Capacity_Factor':'Onshore wind, offshore wind, and solar PV capacity factor',
                            'Power': 'Electricity demand',
                            'Heat': 'Hot water and space heating demand'}
    for data_component, decription in data_components_list.items():
        self.add(data_component)
        print('  - {} variables added'.format(decription))

    print('Areas dataset complete')
    # export dataset
    if save==True:
        self.save_dataset(dir_name)

    return self
